@@ -2,33 +2,37 @@ import numpy as np
 from scipy import stats as scp
 import pandas as pd
 
+from ..data.ProcessData import parse_formula
+from ..misc import summary
 
-#TODO: Add Log-Likehood + AIC + BIC
-#TODO: Add dask for GPU usage
+# TODO: Add Log-Likehood + AIC + BIC
+# TODO: Add dask for GPU usage
 
 
 class OLS:
-    
-    def __init__(self, formula, data, fit_intercept=True):
+
+    def __init__(self, formula, data, fit_intercept=False):
         """Ordinary Least Squares regression
-        
-        :param formula:  Regression formula to be run of the form :obj:`y ~ x1 + x2`.
+
+        :param formula: Regression formula to be run of the form :obj:`y ~ x1 + x2`. See :func:`~parse_formula` in `ProcessData <../../data/process.html#statinf.data.ProcessData.parse_formula>`_
         :type formula: :obj:`str`
         :param data: Input data with Pandas format.
         :type data: :obj:`pandas.DataFrame`
-        :param fit_intercept: Used for adding intercept in the regression formula, defaults to True.
+        :param fit_intercept: Used for adding intercept in the regression formula, defaults to False.
         :type fit_intercept: :obj:`bool`, optional
         """
 
         super(OLS, self).__init__()
         # Parse formula
-        self.no_space_formula = formula.replace(' ', '')
-        self.Y_col = self.no_space_formula.split('~')[0]
-        self.X_col = self.no_space_formula.split('~')[1].split('+')
+        df, self.X_col, self.Y_col = parse_formula(data=data, formula=formula, check_values=True, return_all=True)
+        # self.no_space_formula = formula.replace(' ', '')
+        # self.Y_col = self.no_space_formula.split('~')[0]
+        # self.X_col = self.no_space_formula.split('~')[1].split('+')
+        self.formula = formula
         # Subset X
-        self.X = data[self.X_col].to_numpy()
+        self.X = df[self.X_col].to_numpy()
         # Target variable
-        self.Y = data[self.Y_col].to_numpy()
+        self.Y = df[self.Y_col].to_numpy()
         # Degrees of freedom of the population
         self.dft = self.X.shape[0]
         # Degree of freedom of the residuals
@@ -39,13 +43,15 @@ class OLS:
         self.p = self.X.shape[1]
         # Use intercept or only explanatory variables
         self.fit_intercept = fit_intercept
-    
+        # Beta standard errors for confidence intervals
+        self.std_err_beta = None
+
     def _get_X(self):
         if self.fit_intercept:
             return(np.hstack((np.ones((self.n, 1), dtype=self.X.dtype), self.X)))
         else:
             return(self.X)
-    
+
     def get_betas(self):
         """Computes the estimates for each explanatory variable
 
@@ -59,7 +65,7 @@ class OLS:
         XtY = self._get_X().T.dot(self.Y)
         beta = XtX_1.dot(XtY)
         return(beta)
-    
+
     def fitted_values(self):
         """Computes the estimated values of Y
 
@@ -73,7 +79,7 @@ class OLS:
         for i in range(len(betas)):
             Y_hat += (betas[i] * self._get_X().T[i])
         return(Y_hat)
-    
+
     def _get_error(self):
         """Compute the error term/residuals
 
@@ -89,44 +95,44 @@ class OLS:
         """Residual Sum of Squares
 
         :formula: .. math:: RSS = \\sum_{i=1}^{n} (y_{i} - \\hat{y}_{i})^{2}
-            
+
             where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i` and :math:`\\hat{y}_{i}` denotes the predicted value of :math:`y` for individual :math:`i`.
 
         :return: Residual Sum of Squares.
         :rtype: :obj:`float`
         """
         return((self._get_error()**2).sum())
-    
+
     def tss(self):
         """Total Sum of Squares
 
         :formula: .. math:: TSS = \\sum_{i=1}^{n} (y_{i} - \\bar{y})^{2}
-            
+
             where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i` and :math:`\\bar{y}_{i}` denotes the average value of :math:`y`.
 
         :return: Total Sum of Squares.
         :rtype: :obj:`float`
         """
-        
+
         y_bar = self.Y.mean()
         total_squared = (self.Y - y_bar) ** 2
         return(total_squared.sum())
-    
+
     def r_squared(self):
         """:math:`R^{2}` -- Goodness of fit
 
-        :formula: .. math:: R^{2} = 1 - \dfrac{RSS}{TSS}
+        :formula: .. math:: R^{2} = 1 - \\dfrac{RSS}{TSS}
 
         :return: Goodness of fit.
         :rtype: :obj:`float`
         """
 
-        return(1 - self.rss()/self.tss())
+        return(1 - self.rss() / self.tss())
 
     def adjusted_r_squared(self):
         """Adjusted-:math:`R^{2}` -- Goodness of fit
 
-        :formula: .. math:: R^{2}_{adj} = 1 - (1 - R^{2}) \dfrac{n - 1}{n - p - 1}
+        :formula: .. math:: R^{2}_{adj} = 1 - (1 - R^{2}) \\dfrac{n - 1}{n - p - 1}
 
             where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size
 
@@ -138,11 +144,11 @@ class OLS:
 
         adj_r_2 = 1 - (1 - self.r_squared()) * (self.n - 1) / (self.n - self.p - 1)
         return(adj_r_2)
-    
+
     def _fisher(self):
         """Fisher test
 
-        :formula: .. math:: \\mathcal{F} = \dfrac{TSS - RSS}{\\frac{RSS}{n - p}}
+        :formula: .. math:: \\mathcal{F} = \\dfrac{TSS - RSS}{\\frac{RSS}{n - p}}
 
             where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size
 
@@ -153,36 +159,17 @@ class OLS:
         """
         MSE = (self.tss() - self.rss()) / (self.p - 1)
         MSR = self.rss() / self.dfe
-        return(MSE/MSR)
-    
-    """
-        Returns statistics summary for estimates
-        
-        Formula
-        -------
-        The p-values are computes as:
-        p_value = 2 * (1 - T_n(t_value))
-        
-         * T denotes the Student Cumulative Distribution Function with n degrees of freedom
-        
-        The covariance matrix of beta is compute by:
-        Var(b) = sigma_b**2 * (X'X)^-1
-        
-         * sigma_b denotes the standard deviation computed for a given estimate b
-                
-        Reference
-        ---------
-        
-        """
+        return(MSE / MSR)
+
     def summary(self, return_df=False):
         """Statistical summary for OLS
-        
-        :param return_df: Return the summary as a Pandas DataFrame, else print a string, defaults to False.
+
+        :param return_df: Return the summary as a Pandas DataFrame, else returns a string, defaults to False.
         :type return_df: :obj:`bool`
 
         :formulae: * Fisher test:
 
-            .. math:: \\mathcal{F} = \dfrac{TSS - RSS}{\\frac{RSS}{n - p}}
+            .. math:: \\mathcal{F} = \\dfrac{TSS - RSS}{\\frac{RSS}{n - p}}
 
             where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size
 
@@ -204,57 +191,87 @@ class OLS:
         :references: * Student. (1908). The probable error of a mean. Biometrika, 1-25.
             * Shen, Q., & Faraway, J. (2004). `An F test for linear models with functional responses <https://www.jstor.org/stable/24307230>`_. Statistica Sinica, 1239-1257.
             * Wooldridge, J. M. (2016). `Introductory econometrics: A modern approach <https://faculty.arts.ubc.ca/nfortin/econ495/IntroductoryEconometrics_AModernApproach_FourthEdition_Jeffrey_Wooldridge.pdf>`_. Nelson Education.
+
+        :return: Model's summary.
+        :rtype: :obj:`pandas.DataFrame` or :obj:`str`
         """
         # Initialize
         betas = self.get_betas()
         X = self._get_X()
 
-        sigma_2 = (sum((self._get_error())**2))/(len(X) - len(X[0]))
+        sigma_2 = (sum((self._get_error())**2)) / (len(X) - len(X[0]))
         variance_beta = sigma_2 * (np.linalg.inv(np.dot(X.T, X)).diagonal())
-        std_err_beta = np.sqrt(variance_beta)
-        t_values = betas / std_err_beta
-        
-        p_values = [2 * (1 - scp.t.cdf(np.abs(i), (len(X)-1))) for i in t_values]
-        
+        self.std_err_beta = np.sqrt(variance_beta)
+        t_values = betas / self.std_err_beta
+
+        p_values = [2 * (1 - scp.t.cdf(np.abs(i), (len(X) - 1))) for i in t_values]
+
         summary_df = pd.DataFrame()
         summary_df["Variables"] = ['(Intercept)'] + self.X_col if self.fit_intercept else self.X_col
         summary_df["Coefficients"] = betas
-        summary_df["Standard Errors"] = std_err_beta
+        summary_df["Standard Errors"] = self.std_err_beta
         summary_df["t-values"] = t_values
         summary_df["Probabilities"] = p_values
-        
-        r2 = self.r_squared()
-        adj_r2 = self.adjusted_r_squared()
+
+        _r2 = round(self.r_squared(), 5)
+        ar2 = round(self.adjusted_r_squared(), 5)
+        _n_ = self.n
+        _p_ = self.p
         #
-        fisher = self._fisher()
+        fis = round(self._fisher(), 3)
         #
         if return_df:
             return(summary_df)
         else:
-            print('=========================================================================')
-            print('                               OLS summary                               ')
-            print('=========================================================================')
-            print('| R² = {:.5f}                  | Adjusted-R² = {:.5f}'.format(r2, adj_r2))
-            print('| n  = {:6}                   | p = {:5}'.format(self.n, self.p))
-            print('| Fisher = {:.5f}                         '.format(fisher))
-            print('=========================================================================')
-            print(summary_df.to_string(index=False))
+            summ = f"==================================================================================\n"
+            summ += f'|                                  OLS summary                                   |\n'
+            summ += f"==================================================================================\n"
+            summ += f"| McFadden's R²  =         {_r2:10} | McFadden's R² Adj.  =         {ar2:10} |\n"
+            summ += f"| n              =         {_n_:10} | p                   =         {_p_:10} |\n"
+            summ += f"| Fisher value   =    {fis:15} |                                          |\n"
+            summ += f"==================================================================================\n"
+            summ += summary(summary_df)
+            return(summ)
 
-    def predict(self, new_data):
+    def predict(self, new_data, conf_level=None):
         """Predicted :math:`\\hat{Y}` values for for a new dataset
-        
+
         :param new_data: New data to evaluate with pandas data-frame format.
         :type new_data: :obj:`pandas.DataFrame`
-        
-        :formula: .. math:: \\hat{Y} = X \\hat{\\beta}
+        :param conf_level: Level of the confidence interval, defaults to None.
+        :type conf_level: :obj:`float`
 
-        :return: Predictions
+        :formulae: .. math:: \\hat{Y} = X \\hat{\\beta}
+
+        The confidence interval is computed as:
+
+            .. math:: \\left[ \\hat{Y} \\pm z_{1 - \\frac{\\alpha}{2}} \\dfrac{\\sigma}{\\sqrt{n - 1}} \\right]
+
+        :return: Predictions :math:`\\hat{Y}`
         :rtype: :obj:`numpy.array`
         """
-        X_array = new_data[self.X_col].to_numpy()
+        df = parse_formula(data=new_data, formula=self.formula, check_values=True)
+        X_array = df[self.X_col].to_numpy()
         if self.fit_intercept:
-            new_X = np.hstack((np.ones((new_data.shape[0], 1), dtype=X_array.dtype), X_array))
+            new_X = np.hstack((np.ones((df.shape[0], 1), dtype=X_array.dtype), X_array))
         else:
             new_X = X_array
-        
-        return np.dot(new_X, self.get_betas())
+
+        y_pred = np.dot(new_X, self.get_betas())
+
+        # No CI required
+        if conf_level is None:
+            pred = y_pred
+        else:
+            assert (conf_level > 0.) & (conf_level < 1.), "Your confidence level needs to be between 0 and 1."
+            # User needs CI
+            if self.std_err_beta is None:
+                _ = self.summary(return_df=True)
+
+            quant_order = 1 - ((1 - conf_level) / 2)
+            cv = scp.norm.ppf(quant_order)
+            pred = pd.DataFrame({'Prediction': y_pred,
+                                 'LowerBound': y_pred - cv * (self.std_err_beta.sum() / np.sqrt(self.n - 1)),
+                                 'UpperBound': y_pred + cv * (self.std_err_beta.sum() / np.sqrt(self.n - 1))})
+
+        return pred
